// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
// f90cxx.h
// 
// Fortran to C++ interface
//
// Copyright (c) 1997-2016 European Space Agency (ESA)

#ifndef f90cxx_h
#define f90cxx_h

#include <vector>
#include <iostream>
#include <cassert>
#include <stdint.h>
//#ifdef GFORTRAN
//#include <libgfortran.h>
//#endif

// Until now we defined here NAGF90 and GNUCC to use them 
// locally within this file. They were never defined in the
// configure. 
//
// However, we want to dfine macros in the configure stage 
// to enable specific pieces of code. 
//
// I/O initialization for NAG Fortran can not be used with GNU Fortran
// and the only macro defined in the SAS configure for NAG Fortran is
// NAGf90Fortran. Therefore we enable these io_init and io_finish
// only for NAGf90Fortran.
//
// NAGf90Fortran is also used in src/cfortran.h.

#ifdef NAGf90Fortran

extern "C" void f90_io_init();
extern "C" void f90_io_finish();

#define f90init f90_io_init
#define f90finish f90_io_finish

#endif 


// For the time being we will assume all next defintions are valid 
// for both NAG Fortran and GNU Fortran.

typedef char Integer1;
typedef short Integer2;
typedef int Integer3;
typedef int Logical;
typedef double Double;
typedef float Single;
typedef char* Pointer;


// On 64 bit kernel the components of 
// Triplet (extent, mult, lower) and the offset
// variable must be all 64 bit integers
//
// We define here new types dope_Offset and dope_Chrlen
// which depend on processor type through
// OBJECT_SIZE_64 and CHRLEN_SIZE_64 preprocessor.
// Then assign these types to extent, mult, lower and offset 
//
// Adapted from NAG dope.h (nagfor 5.2)

#ifdef OBJECT_SIZE_64
typedef long long dope_Offset;
#else
typedef int dope_Offset;
#endif

#if defined(CHRLEN_SIZE_64) && defined(OBJECT_SIZE_64)
typedef long long dope_Chrlen;
#else
typedef int dope_Chrlen;
#endif

#if __GNUC__ < 8   \


/* AI Copied from libgfortran.h */
/* FIXME: Increase to 15 for Fortran 2008. Also needs changes to
   GFC_DTYPE_RANK_MASK. See PR 36825.  */
#define GFC_MAX_DIMENSIONS 7

#define GFC_DTYPE_RANK_MASK 0x07
#define GFC_DTYPE_TYPE_SHIFT 3
#define GFC_DTYPE_TYPE_MASK 0x38
#define GFC_DTYPE_SIZE_SHIFT 6


typedef ptrdiff_t index_type;

struct Triplet {
	index_type _stride;
	index_type lower_bound;
	index_type _ubound;

};

#define GFC_DIMENSION_EXTENT(dim) ((dim)._ubound + 1 - (dim).lower_bound)

template<class T, int N>
struct Dope {

  Dope() : addr(0) {}

  // old Dope
  Dope(const T* vec, const unsigned long* ext, const unsigned long* low=0) {
    addr = const_cast<T*>(vec);		// address to first element

    //Witht the new GCC v.9.2 we had to introduce in dal_implementation a way to
    //identify the sizeof the array descriptor in the C++ side. We use the low
    // parameter. In GCC < v.8.0 this parameter should always be 0
    if (low != 0)
      {
	low = 0;
      }
    offset = 0;
    dtype = 265;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j]._ubound	=  ext[i]-1;
      dim[j].lower_bound	= (low ? low[i] : 0);

      if (j == 0)
        dim[j]._stride =  1;
      else
      {

    	  dim[j]._stride =  dim[j-1]._stride * (dim[j-1]._ubound + 1 -  dim[j-1].lower_bound) ;
      }

      offset -= dim[j]._stride * dim[j].lower_bound;


    }
  }

  Dope(const T* vec, const std::vector<unsigned long>& ext,
	const std::vector<unsigned long>& low=std::vector<unsigned long>( 0) ) {
    assert(ext.size() == N);
    assert(low.size() == 0 || low.size() == N);

    addr = const_cast<T*>(vec);		// address to first element
    offset = 0;
    dtype = 0;

    dtype = dtype | N;

    dtype = dtype | (8 << GFC_DTYPE_TYPE_SHIFT);


    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j]._ubound	= ext[i]-1;
      dim[j].lower_bound	= 0;


      if (j == 0)
        dim[j]._stride = 1;
      else
    	  dim[j]._stride =  dim[j-1]._stride * (dim[j-1]._ubound + 1 -  dim[j-1].lower_bound) ;

      offset -= dim[j]._stride * dim[j].lower_bound;

    }
  }
  //old Dope


  Dope(const T* vec, const uint32_t* ext, const uint32_t* low=0) {
    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j]._ubound	= ext[i]-1;
      dim[j].lower_bound	= (low ? low[i] : 0);

      if (j == 0)
        dim[j]._stride = 1;
      else
    	  dim[j]._stride =  dim[j-1]._stride * (dim[j-1]._ubound + 1 -  dim[j-1].lower_bound);

      offset -= dim[j]._stride * dim[j].lower_bound;

    }
  }

  Dope(const T* vec, const std::vector<uint32_t>& ext,
	const std::vector<uint32_t>& low=std::vector<uint32_t>( 0) ) {
    assert(ext.size() == N);
    assert(low.size() == 0 || low.size() == N);

    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j]._ubound	= ext[i]-1;
      dim[j].lower_bound	= (low.size() > 0 ? low[i] : 0);

      if (j == 0)
        dim[j]._stride = 1;
      else
    	  dim[j]._stride =  dim[j-1]._stride * (dim[j-1]._ubound + 1 -  dim[j-1].lower_bound);

      offset -= dim[j]._stride * dim[j].lower_bound;

    }
  }


  T* addr;
  size_t offset;
  index_type dtype;
  Triplet dim[N];
};

template<class T, int N>
inline std::ostream& operator<<(std::ostream& os, const Dope<T,N>& d)
{
  os <<"addr="<<(void*)d.addr << " offset="<<d.offset;
  for (int i = 0; i < N; ++i)
    os << " _ubound="<<d.dim[i]._ubound << " _stride="<<d.dim[i]._stride << " lower_bound="<<d.dim[i].lower_bound;
  os << std::endl;
  return os;
}

template<int N>
struct ChDope {

  ChDope() : addr(0) {}
  char* addr;
  //dope_Chrlen len;
  size_t offset;
  index_type dtype;
  Triplet dim[N];
};

template<int N>
inline std::ostream& operator<<(std::ostream& os, const ChDope<N>& d)
{
  os <<"addr="<<(void*)d.addr << " offset="<<d.offset;
  for (int i = 0; i < N; ++i)
    os << " _ubound="<<d.dim[i]._ubound << " _stride="<<d.dim[i]._stride << " lower_bound="<<d.dim[i].lower_bound;
  os << std::endl;
  return os;
}


#define F90PROCDEF(ret,proc) extern "C" ret proc##_
#define F90MODPROCDEF(ret,mod,proc) extern "C" ret __##mod##_MOD_##proc
#define F90MODPROCPROTO(ret,mod,proc) extern "C" ret __##mod##_MOD_##proc
#define F90MODPROC(mod,proc) __##mod##_MOD_##proc
#define F90MODVARDEF(type,mod,var) type __##mod##_MOD_##var
#define F90MODVAR(mod,var) __##mod##_MOD_##var


// Backwards compatibility for CAL/OAL

typedef Dope<char,1> Dope1;
typedef Dope<char,2> Dope2;
typedef Dope<char,3> Dope3;
typedef Dope<char,4> Dope4;
typedef Dope<char,5> Dope5;



#else

typedef ptrdiff_t index_type;

typedef struct descriptor_dimension {
  index_type _stride;
  index_type lower_bound;
  index_type _ubound;
}
descriptor_dimension;

typedef struct dtype_type
{
  size_t elem_len;
  int version;
  signed char rank; //int8_t rank;
  signed char type; //int16_t type;
  
}
dtype_type;

#define GFC_MAX_DIMENSIONS 15

#define GFC_DTYPE_RANK_MASK 0x0F
#define GFC_DTYPE_TYPE_SHIFT 4
#define GFC_DTYPE_TYPE_MASK 0x70
#define GFC_DTYPE_SIZE_SHIFT 7

template<class T, int N,int size=0>
struct Dope {

Dope() : addr(0) {}

  // old Dope
  Dope(const T* vec, const unsigned long* ext, const unsigned long* low=0) {
    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;
    low = 0; 
    dtype.rank = N;
    dtype.elem_len = 0;
    dtype.type = 1;
    span = size;
    //cout  << "Dope1  N "<< N << " size "<< size << endl;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j]._ubound	=  ext[i]-1;
      dim[j].lower_bound	= (low ? low[i] : 0);

      if (j == 0)
        dim[j]._stride =  1;
      else      
    	  dim[j]._stride =  dim[j-1]._stride * (dim[j-1]._ubound + 1 -  dim[j-1].lower_bound) ;      

      offset -= dim[j]._stride * dim[j].lower_bound;


    }
  }

  Dope(const T* vec, const std::vector<unsigned long>& ext,
	const std::vector<unsigned long>& low=std::vector<unsigned long>( 0) ) {
    assert(ext.size() == N);
    assert(low.size() == 0 || low.size() == N);

    addr = const_cast<T*>(vec);		// address to first element
    span = size;
    //cout << "Dope2  span " << span << " sizeof(int32_t) " << sizeof(int32_t) 
    //	    << " N "<< N << " size "<< size << endl;

    offset = 0;
    dtype.rank = N;
    dtype.elem_len = 0;
    dtype.type = 1;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j]._ubound	= ext[i]-1;
      dim[j].lower_bound	= 0;


      if (j == 0)
        dim[j]._stride = 1;
      else
    	  dim[j]._stride =  dim[j-1]._stride * (dim[j-1]._ubound + 1 -  dim[j-1].lower_bound) ;

      offset -= dim[j]._stride * dim[j].lower_bound;

    }
  }
  //old Dope


  Dope(const T* vec, const uint32_t* ext, const uint32_t* low=0) {
    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;
    dtype.rank = N;
    dtype.elem_len = 0;
    dtype.type = 1;
    span = size;
    //cout << "Dope3  span " << span << " sizeof(int32_t) " << sizeof(int32_t)
    //        << " N "<< N << " size "<< size << endl;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j]._ubound	= ext[i]-1;
      dim[j].lower_bound	= (low ? low[i] : 0);

      if (j == 0)
        dim[j]._stride = 1;
      else
    	  dim[j]._stride =  dim[j-1]._stride * (dim[j-1]._ubound + 1 -  dim[j-1].lower_bound);

      offset -= dim[j]._stride * dim[j].lower_bound;

    }
  }

  Dope(const T* vec, const std::vector<uint32_t>& ext,
	const std::vector<uint32_t>& low=std::vector<uint32_t>( 0) ) {
    assert(ext.size() == N);
    assert(low.size() == 0 || low.size() == N);

    addr = const_cast<T*>(vec);		// address to first element
    //cout << "Dope4" << endl;

    offset = 0;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j]._ubound	= ext[i]-1;
      dim[j].lower_bound	= (low.size() > 0 ? low[i] : 0);

      if (j == 0)
        dim[j]._stride = 1;
      else
    	  dim[j]._stride =  dim[j-1]._stride * (dim[j-1]._ubound + 1 -  dim[j-1].lower_bound);

      offset -= dim[j]._stride * dim[j].lower_bound;

    }
  }


  T* addr;
  size_t offset;
  dtype_type dtype;
  index_type span;
  descriptor_dimension dim[N];

};

template<class T, int N,int size=0>
inline std::ostream& operator<<(std::ostream& os, const Dope<T,N,size>& d)
{
  os <<"addr="<<(void*)d.addr << " offset="<<d.offset;
  for (int i = 0; i < N; ++i)
    os << " _ubound="<<d.dim[i]._ubound << " _stride="<<d.dim[i]._stride << " lower_bound="<<d.dim[i].lower_bound;
  os << std::endl;
  return os;
}

template<int N>
struct ChDope {

  ChDope() : addr(0) 
  {
    //cout << "ChDope" << endl;
    dtype.rank = 1;
    dtype.elem_len = 0;
    dtype.type = 1;
    span = sizeof(long long);
    offset = 0;

  }
  char* addr;
  //dope_Chrlen len;
  size_t offset;
  dtype_type dtype;
  index_type span;
  descriptor_dimension dim[N];
};

template<int N>
inline std::ostream& operator<<(std::ostream& os, const ChDope<N>& d)
{
  os <<"addr="<<(void*)d.addr << " offset="<<d.offset;
  for (int i = 0; i < N; ++i)
    os << " _ubound="<<d.dim[i]._ubound << " _stride="<<d.dim[i]._stride << " lower_bound="<<d.dim[i].lower_bound;
  os << std::endl;
  return os;
}


#define F90PROCDEF(ret,proc) extern "C" ret proc##_
#define F90MODPROCDEF(ret,mod,proc) extern "C" ret __##mod##_MOD_##proc
#define F90MODPROCPROTO(ret,mod,proc) extern "C" ret __##mod##_MOD_##proc
#define F90MODPROC(mod,proc) __##mod##_MOD_##proc
#define F90MODVARDEF(type,mod,var) type __##mod##_MOD_##var
#define F90MODVAR(mod,var) __##mod##_MOD_##var

// Backwards compatibility for CAL/OAL

typedef Dope<char,1,sizeof(char)> Dope1;
typedef Dope<char,2,sizeof(char)> Dope2;
typedef Dope<char,3,sizeof(char)> Dope3;
typedef Dope<char,4,sizeof(char)> Dope4;
typedef Dope<char,5,sizeof(char)> Dope5;



#endif







/* NAGf90Fortran array descriptor definitions */

#ifdef NAGf90Fortran
struct Triplet {
  dope_Offset extent,mult,lower;

};

template<class T, int N>
struct Dope {

  Dope() : addr(0) {}

  // old Dope
  Dope(const T* vec, const unsigned long* ext, const unsigned long* low=0) {
    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;


    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j].extent	= ext[i];
      dim[j].lower	= (low ? low[i] : 0);

      if (j == 0)
        dim[j].mult = sizeof(T);
      else
        dim[j].mult = dim[j-1].mult * dim[j-1].extent;

      offset -= dim[j].mult * dim[j].lower;

    }
  }

  Dope(const T* vec, const std::vector<unsigned long>& ext,
	const std::vector<unsigned long>& low=std::vector<unsigned long>( 0) ) {
    assert(ext.size() == N);
    assert(low.size() == 0 || low.size() == N);

    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j].extent	= ext[i];
      dim[j].lower	= (low.size() > 0 ? low[i] : 0);

      if (j == 0)
        dim[j].mult = sizeof(T);
      else
        dim[j].mult = dim[j-1].mult * dim[j-1].extent;

      offset -= dim[j].mult * dim[j].lower;

    }
  }
  //old Dope


  Dope(const T* vec, const uint32_t* ext, const uint32_t* low=0) {
    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;
    
    
    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j].extent	= ext[i];
      dim[j].lower	= (low ? low[i] : 0);

      if (j == 0)
        dim[j].mult = sizeof(T);
      else
        dim[j].mult = dim[j-1].mult * dim[j-1].extent;

      offset -= dim[j].mult * dim[j].lower;
      
    }
  }

  Dope(const T* vec, const std::vector<uint32_t>& ext,
	const std::vector<uint32_t>& low=std::vector<uint32_t>( 0) ) {
    assert(ext.size() == N);
    assert(low.size() == 0 || low.size() == N);

    addr = const_cast<T*>(vec);		// address to first element

    offset = 0;

    // Swap C++ and Fortran indices
    for (int i = N-1, j = 0; i >= 0; --i, ++j) {
      dim[j].extent	= ext[i];
      dim[j].lower	= (low.size() > 0 ? low[i] : 0);

      if (j == 0)
        dim[j].mult = sizeof(T);
      else
        dim[j].mult = dim[j-1].mult * dim[j-1].extent;

      offset -= dim[j].mult * dim[j].lower;

    }
  }


  T* addr;
  dope_Offset offset;
  Triplet dim[N];
};

template<class T, int N>
inline std::ostream& operator<<(std::ostream& os, const Dope<T,N>& d)
{
  os <<"addr="<<(void*)d.addr << " offset="<<d.offset;
  for (int i = 0; i < N; ++i)
    os << " extent="<<d.dim[i].extent << " mult="<<d.dim[i].mult << " lower="<<d.dim[i].lower;
  os << std::endl;
  return os;
}

template<int N>
struct ChDope {

  ChDope() : addr(0) {}
  char* addr;
  dope_Chrlen len;
  dope_Offset offset;
  Triplet dim[N];
};

template<int N>
inline std::ostream& operator<<(std::ostream& os, const ChDope<N>& d)
{
  os <<"addr="<<(void*)d.addr << " offset="<<d.offset;
  for (int i = 0; i < N; ++i)
    os << " extent="<<d.dim[i].extent << " mult="<<d.dim[i].mult << " lower="<<d.dim[i].lower;
  os << std::endl;
  return os;
}


#define F90PROCDEF(ret,proc) extern "C" ret proc##_
#define F90MODPROCDEF(ret,mod,proc) extern "C" ret mod##_MP_##proc
#define F90MODPROCPROTO(ret,mod,proc) extern "C" ret mod##_MP_##proc
#define F90MODPROC(mod,proc) mod##_MP_##proc
#define F90MODVARDEF(type,mod,var) type mod##_MP_##var
#define F90MODVAR(mod,var) mod##_MP_##var

// Backwards compatibility for CAL/OAL

typedef Dope<char,1> Dope1;
typedef Dope<char,2> Dope2;
typedef Dope<char,3> Dope3;
typedef Dope<char,4> Dope4;
typedef Dope<char,5> Dope5;



#endif


#endif
