// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

#ifndef __stringutil_h
#define __stringutil_h

#include <string>
#include <Tracer.h>
#include <sstream>
#include <iomanip>
#include <vector>
#include <utility>
#include <strstream>
#ifdef __APPLE__
 #include <unistd.h>
#endif


std::string trim(const std::string& s, bool leading=false,
				 const std::string &ws=" \n\t");

std::string trimDigits(const std::string &s);

std::string unquote(const std::string &);

std::string lowercase(const std::string &);

std::string uppercase(const std::string&);

template<class T> std::string toString(const T& t, int prec = -1)
{
	PROC
	std::stringstream s;
	if (prec >= 0)
		s.precision(prec);
	s<<t;

	return s.str();
}

template<class T> T toNumber(const std::string &s)
{
	PROC

	std::stringstream str;
	str<<s;
	T v;	str>>v;
	if(str.fail()){
	  std::string ss("toNumber failed: [" + s + "] conversion failed");
	  throw(ss);
	}
	RETURN v;
}


//
//	replace all occurences of a string within a string by another string
//	replaceall("garbage in garbage out", "garbage", "beauty") ==
//		"beauty in beauty out"
//
std::string replaceall(const std::string &str,
					   const std::string &searchStr,
					   const std::string &replStr);


//
//	quotedBreakup: break a string up into tokens separated by 
//	white-space characters (along the lines of std::strtok());
//	the delimiters can also contain open parenthesis "({["
//	in which case the corresponding closing delimiter is
//	taken as the associated closing parenthesis
//	breakup		 : same as quotedBreakup() but quoting characters are
//				   ignored
//
//		||			=> {}
//		|:|			=> {}
//		|"|			=> {"}
//		|s1|		=> {s1}
//		|s1"::"x|	=> {s1"::"x}
//		|s1:s2:s"3|	=> {s1,s2,s"3}
//		|s1:s2:s3|	=> {s1,s2,s3}
//		|"x::":s1|	=> {"x::",s1}
//		|x(1, 2) x={x 2}  => {x(1, 2), x, {2 3}} with quote="({"
//
//	UL, Wed Jun  3 13:57:07 MET DST 1998
//
//
std::vector<std::string> quotedBreakup(const std::string &str,
									   const std::string &sep = "\t\n ,;:",
									   const std::string &quote = "\"\'");

std::vector<std::string> breakup(const std::string &str,
								 const std::string &sep = "\t\n ,;:");

//
//	fullbreakup: break a string up into tokens separated by white-space
//			 characters - allowing for null size tokens
//
//
std::vector<std::string> fullbreakup(const std::string &str,
									 const std::string &sep = ":" );


//
//	buildup: construct a string by concatenating the list of
//      separated tokens.  The separator is specified with sep.
//
std::string buildup( const std::vector<std::string>& tokens, char sep = ':' );


//
//	removeStringToken: Removes the specified token (the first
//                   token has position zero) token from the given
//                   list of separated tokens, and returns the result.
//                   The separator is given by sep.
//
std::string removeStringToken(const std::string &tokens,
							  unsigned int position = 0,
							  char sep = ':' );


//
//	stringToken: Return the specified (the first token has position zero)
//                  token from the given list of separated tokens. 
//                  The separator is given by sep.
//
std::string stringToken(const std::string &tokens,
						unsigned int position = 0,
						char sep = ':' );

//
//	isBlank: Returns true if the given string contains
//	only blank (ASCII 32) characters, or is empty.
//
bool isBlank(const std::string &thisString);

//
//	utility functions related to file names
//
namespace Filename {
	//	split file name into directory + base name; everything after a ':'
	//	gets chopped off the basename (DAL compound syntax)
	//
	//	"/etc/passwd"   -> "/etc" + "passwd"
	//	"/"				-> "/" + ""
	//	""				-> "" + ""
	//	"."				-> "." + ""
	//	"./"			-> "." + ""
	//	"./bla.txt"		-> "." + "bla.txt"
	//	"bla.txt"		-> "." + "bla.txt"
	//	"bla.txt:TABLE%ATTRIBUTE -> "." + "bla.txt"
	//
	typedef std::pair<std::string, std::string> StringPair;
	StringPair split(const std::string &name);

	//	Is a given name a directory name?
	bool isDirectoryName(const std::string &name);

        string make_temp_filename(const std::string &name_template);
        string make_temp_filename(const std::string &prefix,
                                     const std::string &suffix);
}

#endif
