// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.

/* Revision History
 *
 * $Id: token.h,v 1.18 2008/09/26 16:38:55 eojero Exp $
 *
 */


/**************************************************************************/
/*                                                                        */
/*  FILENAME                                                              */
/*  ========                                                              */
/*  token.h                                                               */
/*                                                                        */
/*                                                                        */
/*  SUBSYSTEM                                                             */
/*  =========                                                             */
/*  XMM SAS -  UTILILITES                                                 */
/*                                                                        */
/*                                                                        */
/*  TITLE                                                                 */
/*  =====                                                                 */
/*  Public header file for interface to TokenT type.                      */
/*                                                                        */
/*                                                                        */
/*  AUTHORS                                                               */
/*  =======                                                               */
/*  Mark Thomas                                                           */
/*  CARA Software & Services Ltd                                          */
/*                                                                        */
/*                                                                        */
/*  CONSTRAINTS                                                           */
/*  ===========                                                           */
/*  All typedefs and functions given in this header file are for          */
/*  public use.                                                           */
/*                                                                        */
/*                                                                        */
/*  DESCRIPTION                                                           */
/*  ===========                                                           */
/*  Public Header file for interface to TokenT type.                      */
/*                                                                        */
/*                                                                        */
/*  EXAMPLE                                                               */
/*  =======                                                               */
/*                                                                        */
/*                                                                        */
/*  #include "token.h"                                                    */
/*                                                                        */
/*                                                                        */
/*  main()                                                                */
/*  {                                                                     */
/*  	TokenT  Token;                                                    */
/*                                                                        */
/*                                                                        */
/*      Token.Set( LITERAL_STRING, "This is a string" );                  */
/*                                                                        */
/*                                                                        */
/*  }                                                                     */
/*                                                                        */
/*                                                                        */
/**************************************************************************/


#ifndef __TOKEN_H
#define __TOKEN_H


#include <string>
#include <iostream>
#include <vector>

#include "Vec.h"


#include "binary_tree.h"

class TokenValue;

class Functor;
class PointerDescriptor;

class Token
{
	public :

		typedef BinaryTree<Token>	BTree;

		enum IDType
		{
		/*	ID			     	   TYPE                 */
		/*	========================================================*/

			EndOfFile,		// none 
			LessEqual,		// none
			GreaterEqual,		// none
			EqualEqual,		// none
			NotEqual,		// none
			Defun,			// BTree
			If,			// vector<BTree>
			Loop,			// vector<BTree>
			Exitif,			// BTree
			Return,			// BTree
			PlusPlus,		// none
			MinusMinus,		// none
			Identifier,    	      	// string
			Integer,		// int
			Short,			// short
			Boolean,		// bool
			Real,			// double
			Single,			// float
			String, 	     	// string
			QuotedCharacter,   	// char
			UnsignedInteger,	// unsigned int
			Character,   		// char
			FreeText,	   	// char *
			TokenPointer,		// Token *
			BinaryTreeVector,	// vector<BinaryTree>
			Vector,			// vector<Token>
			Pointer,                // void *
			Ptr,			// PointerDescriptor
			Function,		// Functor
			Undefined		// none

		};


		// Constructors
		Token();
		Token( IDType id );
		Token( int value );
		Token( short value );
		Token( bool value );
		Token( const int * addressOfTheValue );
		Token( unsigned int value );
		//Token( const unsigned int * addressOfTheValue );
		Token( double value );
		Token( float value );
		Token( const double * addressOfTheValue );
		Token( char value, IDType id = Token::Character );
		Token( const char * addressOfTheValue, IDType id );
		Token( const string & value, IDType id = Token::String );
		Token( const Functor & value );
		Token( const PointerDescriptor & value );
		Token( const void * value, IDType id = Token::Pointer );
		Token( const vector<BTree> & value, IDType id );
		Token( const BinaryTree<Token> & value, IDType id );
		Token( const Vec<Token>  & value );
		Token( const Token & );
		Token & operator=( const Token & );
		~Token();

		Token & operator=( IDType id );
		Token & operator=( int value );
		Token & operator=( short value );
		Token & operator=( bool value );
		Token & operator=( double value );
		Token & operator=( float value );
		Token & operator=( const string & value );
		Token & operator=( const void * value );
		Token & operator=( char value );
		Token & operator=( const vector<BTree> & value );
		Token & operator=( const BTree & value );
		Token & operator=( const Vec<Token> & value );
		Token & operator=( const Functor & value );
		Token & operator=( const PointerDescriptor & value );

		void		set( const Token &	usingValueOfThisToken );


		char				asChar() const;
		const char *			asCharPointer() const;
		unsigned int			asUInt() const;
		int				asInt() const;
		short				asShort() const;
		bool				asBool() const;
		double				asReal() const;
		float				asSingle() const;
		char				asCharacter() const;
		string				asString() const;
		const void *			asPointer() const;
		Token				asToken() const;
		Functor				asFunctor() const;
		PointerDescriptor		asPointerDescriptor() const;
		const vector<BTree> &		asBTreeVector() const;
		const BTree &			asBTree() const;
		const Vec<Token> &		asVector() const;

		void		setAddress( const Token * toThis );
		const Token *	addressOf() const;
		Token *		addressOf();
		void		setAddressOfValue( const int * toThisAddress );
		void		setAddressOfValue( const unsigned int * toThisAddress );
		void		setAddressOfValue( const double * toThisAddress );
		void		setAddressOfValue( const string * toThisAddress );
		const void *	addressOfValue() const;
		IDType 		id() const;
		void		setRawValue( const string & rawValue );
		string		rawValue() const;
		void		setDebug( const string & debug );
		string		debug() const;

		Token operator<( const Token & ) const;
		Token operator<=( const Token & ) const;
		Token operator>( const Token & ) const;
		Token operator>=( const Token & ) const;
		Token operator==( const Token & ) const;
		Token operator!=( const Token & ) const;

		Token operator+( const Token & ) const;
		Token operator-( const Token & ) const;
		Token operator*( const Token & ) const;
		Token operator/( const Token & ) const;

		void printOn( ostream &	thisStream ) const;

	private :

		IDType		_id;
		string		_rawValue;
		TokenValue *	_value;
		const Token *	_address;
		string		_debug;
};


inline ostream& operator << (
			ostream &	os,
			const Token &	token )
{ token.printOn( os ); return os; }


class PointerDescriptor
{
	public :
		PointerDescriptor(
			Token::IDType	type,
			unsigned int	indirections,
			const void *	address );

		PointerDescriptor( const PointerDescriptor & value );

		PointerDescriptor & operator=( const PointerDescriptor & value );
		Token::IDType	type() const;
		unsigned int indirections() const;
		const void * raw() const;
		void printOn( ostream & os ) const;

	private :
		Token::IDType		_type;
		unsigned int		_indirections;
		const void *		_address;
};


class Functor
{
	public :

		enum Type { C, Perl, Internal };
		Functor(
			Type			type,
			const void * 		address,
			const Token &		returnType );

		Functor( const Functor & value );
		Functor& operator=( const Functor & value );
		Type type() const;
		const void * address() const;
		Token returnType() const;
		void printOn( ostream & os ) const;

	private :
		Type			_type;
		const void *		_address;
		Token			_returnType;
};


#endif

